/* myRGB2HSV.cpp

J. Colineau 13/12/2007

tir de OpenCV/tests/aColor
adaptation  images 8U

voir aussi http://www.cs.rit.edu/~ncs/color/t_convert.html
*/
#include <math.h>
#include <float.h>
#include "cv.h"
#include "myRGB2HSV.h"

// conversion une ligne
void convert_row_bgr2abc_32f_c3( const float* src_row, float* dst_row, int n )
{
    int depth = CV_8U; //= CV_MAT_DEPTH(test_mat[INPUT][0].type);
    float h_scale = depth == CV_8U ? 30.f : 60.f;
    float scale = depth == CV_8U ? 255.f : depth == CV_16U ? 65535.f : 1.f;
    int j;

    for( j = 0; j < n*3; j += 3 )
    {
        float r = src_row[j+2];
        float g = src_row[j+1];
        float b = src_row[j];
        float vmin = MIN(r,g);
        float v = MAX(r,g);
        float s, h, diff;
        vmin = MIN(vmin,b);
        v = MAX(v,b);
        diff = v - vmin;
        if( diff == 0 )
            s = h = 0;
        else
        {
            s = diff/(v + FLT_EPSILON);
            diff = 1.f/diff;

            h = r == v ? (g - b)*diff :
                g == v ? 2 + (b - r)*diff : 4 + (r - g)*diff;

            if( h < 0 )
                h += 6;
        }

        dst_row[j] = h*h_scale;
        dst_row[j+1] = s*scale;
        dst_row[j+2] = v*scale;
    }
}

// taken from http://www.cs.rit.edu/~ncs/color/t_convert.html
void convert_row_abc2bgr_32f_c3( const float* src_row, float* dst_row, int n )
{
    int depth = CV_8U ;//CV_MAT_DEPTH(test_mat[INPUT][0].type);
    float h_scale = depth == CV_8U ? 1.f/30 : 1.f/60;
    float scale = depth == CV_8U ? 1.f/255 : depth == CV_16U ? 1.f/65535 : 1;
    int j;

    for( j = 0; j < n*3; j += 3 )
    {
        float h = src_row[j]*h_scale;
        float s = src_row[j+1]*scale;
        float v = src_row[j+2]*scale;
        float r = v, g = v, b = v;

        if( h < 0 )
            h += 6;
        else if( h >= 6 )
            h -= 6;

        if( s != 0 )
        {
            int i = cvFloor(h);
            float f = h - i;
            float p = v*(1 - s);
            float q = v*(1 - s*f);
            float t = v*(1 - s*(1 - f));

            if( i == 0 )
                r = v, g = t, b = p;
            else if( i == 1 )
                r = q, g = v, b = p;
            else if( i == 2 )
                r = p, g = v, b = t;
            else if( i == 3 )
                r = p, g = q, b = v;
            else if( i == 4 )
                r = t, g = p, b = v;
            else
                r = v, g = p, b = q;
        }

        dst_row[j] = b;
        dst_row[j+1] = g;
        dst_row[j+2] = r;
    }
}

void RGB2HSV( const IplImage* src, IplImage* dst )
{
	const float c8u = 0.0039215686274509803f; // 1./255
	int depth = src->depth;
	int cn = src->nChannels, dst_cn = dst->nChannels; // ???
	int cols = src->width, dst_cols_n = dst->width*dst_cn;
	float* src_buf = (float*)cvAlloc( src->width*3*sizeof(src_buf[0]) );
    float* dst_buf = (float*)cvAlloc( dst->width*3*sizeof(dst_buf[0]) );
    int i, j;

	assert( (cn == 3 || cn == 4) && (dst_cn == 3 || dst_cn == 1) );
	
	for( i = 0; i < src->height; i++ )
    {
		const uchar* src_row = (const uchar*)(src->imageData + i*src->widthStep);
		uchar* dst_row = (uchar*)(dst->imageData  + i*dst->widthStep );

        for( j = 0; j < cols; j++ )
        {
			// attention: dpend de l'ordre : conversion bgr -> rgb
            src_buf[j*3] = src_row[j*cn + 2]*c8u;
            src_buf[j*3+1] = src_row[j*cn + 1]*c8u;
            src_buf[j*3+2] = src_row[j*cn ]*c8u;
        }

        convert_row_bgr2abc_32f_c3( src_buf, dst_buf, cols );

        for( j = 0; j < dst_cols_n; j++ )
        {
            int t = cvRound( dst_buf[j] );
            dst_row[j] = CV_CAST_8U(t);
        }
     }
	cvFree( (void**) &src_buf );
    cvFree( (void**) &dst_buf );
}

void HSV2RGB( const IplImage* src, IplImage* dst)
{
	// attention, il y a des modifs par rapport au code d'origine 
	//(qui faisait la conversion inverse de dst vers src)
	const float c8u = 0.0039215686274509803f; // 1./255
	int depth = src->depth;

	int src_cn = src->nChannels, cn = dst->nChannels;
    int cols_n = src->width*src_cn, src_cols = src->width;
    float* src_buf = (float*)cvAlloc( src->width*3*sizeof(src_buf[0]) );
    float* dst_buf = (float*)cvAlloc( dst->width*3*sizeof(dst_buf[0]) );
    int i, j;
	assert( cn == 3 || cn == 4 );
	  for( i = 0; i < src->height; i++ )
        {
			const uchar* src_row = (const uchar*)(src->imageData + i*src->widthStep);
			uchar* dst_row = (uchar*)(dst->imageData + i*dst->widthStep);

                    for( j = 0; j < cols_n; j++ )
                        src_buf[j] = src_row[j];

                    convert_row_abc2bgr_32f_c3( src_buf, dst_buf, src_cols );

                    for( j = 0; j < src_cols; j++ )
                    {
                        int b = cvRound( dst_buf[j*3]*255. );
                        int g = cvRound( dst_buf[j*3+1]*255. );
                        int r = cvRound( dst_buf[j*3+2]*255. );
						// attention: dpend de l'ordre : conversion bgr -> rgb
                        dst_row[j*cn + 2] = CV_CAST_8U(b);
                        dst_row[j*cn + 1] = CV_CAST_8U(g);
                        dst_row[j*cn ] = CV_CAST_8U(r);
                        if( cn == 4 )
                            dst_row[j*cn + 3] = 0;
                    }
                }
        cvFree((void**) &src_buf );
        cvFree((void**) &dst_buf );
}


